/* ------------------------------------------------------------------------ *
 *                                                                          *
 * ntp-ssl-demo-pass.c  username/password authenticated                     *
 *                                                                          *
 * This program demonstrates the communication with a Nessus daemon         *
 * and  counts the returned nessus server settings to the screen.           *
 * Unlike ntp-demo.c, it uses SSL for encryption (relies on OpenSSL         *
 * libraries to provide the required en/decryption functions). No client    *
 * certs are used, and ssl_version=SSLv3 must be set in nessusd.conf.       *
 *                                                                          *
 * The Following settings according to NTP 1.2 are retrieved:               *
 * SERVER <|> PLUGIN_LIST <|> [data] <|> SERVER                             *
 * SERVER <|> PREFERENCES <|> [data] <|> SERVER                             *
 * SERVER <|> RULES <|> [data] <|> SERVER                                   *
 *                                                                          *
 * this program has been written and tested with nessusd version 2.2.0      *
 * 20041129 frank4dd                                                        *
 *                                                                          *
 * compile instructions:                                                    *
 * gcc -lssl -lcrypt ntp-ssl-demo-pass.c -o ntp-ssl-demo-pass               *
 *                                                                          *
 * Don't forget to add a user to Nessus, ie. like below:                    *
 *                                                                          *
 * susie:/home/nessus/sbin # ./nessus-adduser                               *
 * Using /var/tmp as a temporary file holder                                *
 *                                                                          *
 * Add a new nessusd user                                                   *
 * ----------------------                                                   *
 *                                                                          *
 * Login : fm2                                                              *
 * Authentication (pass/cert) [pass] :                                      *
 * Login password : test                                                    * 
 * Login password (again) : test                                            *
 *                                                                          *
 * User rules                                                               *
 * ----------                                                               *
 * nessusd has a rules system which allows you to restrict the hosts        *
 * that fm2 has the right to test. For instance, you may want               *
 * him to be able to scan his own host only.                                *
 *                                                                          *
 * Please see the nessus-adduser(8) man page for the rules syntax           *
 *                                                                          *
 * Enter the rules for this user, and hit ctrl-D once you are done :        *
 * (the user can have an empty rules set)                                   *
 *                                                                          *
 * Login             : fm2                                                  *
 * Password          : ***********                                          *
 * DN                :                                                      *
 * Rules             :                                                      *
 *                                                                          *
 * Is that ok ? (y/n) [y]                                                   *
 * user added.                                                              *
 * ------------------------------------------------------------------------ */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <openssl/x509v3.h>
#include <openssl/ssl.h>

/* DEBUG = 0: connects and returns number of plugins, preferences, rules */
/* DEBUG = 1: adds the NTP protocol strings received */
/* DEBUG = 2: shows in addition the received data and counters */
#define DEBUG 2

#define CIPHER_LIST "EDH-RSA-DES-CBC3-SHA:EDH-DSS-DES-CBC3-SHA:DES-CBC3-SHA:DES-CBC3-MD5:DHE-DSS-RC4-SHA:IDEA-CBC-SHA:RC4-SHA:RC4-MD5:IDEA-CBC-MD5:RC2-CBC-MD5:RC4-MD5"

#define DEST_IP "172.20.1.101"
#define DEST_PORT 1241

/* MAXDATASIZE should be large enough to handle a huge plugin *
 * description so we get a newline. If it is to small, we get *
 * chunks with no newline counted and its just a mess.        */
#define MAXDATASIZE 16384

#define USERPROMPT "User : "
#define PASSPROMPT "Password : "
#define USERNAME "fm2"
#define PASSWORD "test"
#define CLIENT_NTP_VERSION "< NTP/1.2 >"
#define SERVER_NTP_VERSION "< NTP/1.2 >"
#define PLUGS_START "SERVER <|> PLUGIN_LIST <|>"
#define PREFS_START "SERVER <|> PREFERENCES <|>"
#define RULES_START "SERVER <|> RULES <|>"
#define SERVER_END_MARKER "<|> SERVER"

char server_ntp_version[255];

int  plugs_counter = 0;
int  prefs_counter = 0;
int  rules_counter = 0;
SSL           *ssl;


/* ------------------------------------------------------------------------- *
 * nessus_connect creates a new tcp session to the nessus server and returns *
 * the file descriptor to it.                                                *
 * ------------------------------------------------------------------------- */
int nessus_connect(char * nessus_ip, int nessus_port) {
  int sockfd;
  struct sockaddr_in dest_addr;
  static SSL_CTX        *ssl_ctx = NULL;
  X509			*servercert;

  if(DEBUG) printf("Creating socket file descriptor.\n");
  sockfd = socket(AF_INET, SOCK_STREAM, 0);

  if(DEBUG) printf("Loading content into filedescriptor.\n");
  dest_addr.sin_family=AF_INET;
  dest_addr.sin_port=htons(nessus_port);
  dest_addr.sin_addr.s_addr=inet_addr(nessus_ip);

  /* Zeroing the rest of the struct */

  memset(&(dest_addr.sin_zero), '\0', 8);

  if(DEBUG) printf("Try connecting to %s.\n", DEST_IP);

  if ( connect(sockfd, (struct sockaddr *) &dest_addr,
                              sizeof(struct sockaddr)) == -1 ) {
    if(DEBUG) perror("Error connecting");
    return -1;
  } else {
    if(DEBUG) printf("Connected to: %s.\n", DEST_IP);
  }

  ssl = NULL;

  /* initialize SSL library and register algorithms */
  if(SSL_library_init() < 0)
    printf("Could not initialize the OpenSSL library !\n");

  /* load  the SSL error messages */
  SSL_load_error_strings();

  /* create a new SSL_CTX object as framework for TLS/SSL enabled functions */
  if((ssl_ctx = SSL_CTX_new(SSLv3_client_method())) == NULL) 
    printf("SSL_CTX_new() context creation error\n");

  /* enable all SSL engine bug workaround options (i.e. Netscape, Microsoft) */
  if(SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL) < 0)
    printf("SSL_CTX_set_options error\n");

  /* choose list of available SSL_CIPHERs with a control string */
  if(! SSL_CTX_set_cipher_list(ssl_ctx, CIPHER_LIST))
    printf("SSL_CTX_set_cipher_list error\n");

  /* create a new SSL structure for a connection */
  if((ssl = SSL_new(ssl_ctx)) == NULL)
     printf("SSL_new() general error\n");

  /* connect the SSL object with the socket file descriptor */
  if(! SSL_set_fd(ssl, sockfd))
     printf("SSL_set_fd() error connecting to socket.\n");

  /* initiate the TLS/SSL handshake with an TLS/SSL server */
  if(SSL_connect(ssl) <= 0)
     printf("SSL_connect() error during SSL handshake\n");

  /* show the connection cipher that is used */
  if(DEBUG) printf("SSL_get_cipher = %s\n", SSL_get_cipher(ssl));

  /* get the received nessus server certificate */
  if(! (servercert = SSL_get_peer_certificate(ssl)))
    printf("SSL_get_peer_certificate() error: cannot get server certificate\n");

  /* print the received nessus server certificate */
  if(DEBUG == 2) {
    BIO                  *outbio;
    outbio = BIO_new_fd(fileno(stdout), BIO_NOCLOSE);
    if (! (X509_print_ex(outbio, servercert, 0, XN_FLAG_SEP_MULTILINE)))
      BIO_printf(outbio, "Error printing certificate text information\n");
  }

  return 1;
}

/* ------------------------------------------------------------------------- *
 * nessus_login tries to log in with the username and password provided. It  *
 * returns 1 for success, 0 for failure.                                     *
 * ------------------------------------------------------------------------- */

int nessus_login(char * username, char * password){

  char * buf = NULL;
  int login = 0 , retcode = 0;
  const char newline[2] = {'\n', '\0'};

  /* send the client protocol version, followed by a newline */
  if(DEBUG) printf("Sending Data: %s%c", CLIENT_NTP_VERSION, '\n');
  retcode = SSL_write(ssl, CLIENT_NTP_VERSION, strlen(CLIENT_NTP_VERSION));
  retcode = SSL_write(ssl, newline, 1);

  buf = (char *) malloc(MAXDATASIZE);

  /* receive the server protocol version, followed by a newline */

  retcode=SSL_read(ssl, buf, strlen(SERVER_NTP_VERSION)+1);
  if(DEBUG) printf("Receive Data: %s", buf);

  /* set global variable server protocol version, remove the newline */
  strncpy(server_ntp_version, buf, retcode-1);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  /* receive the username prompt */
  retcode=SSL_read(ssl, buf, MAXDATASIZE);
  if(DEBUG) printf("Receive Data: %s\n", buf);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  /* send the username followed by a newline */
  if(DEBUG) printf("Sending Data: %s%c", USERNAME, '\n');
  retcode = SSL_write(ssl, USERNAME, strlen(USERNAME));
  retcode = SSL_write(ssl, newline, 1);

  /* receive the password prompt */
  retcode=SSL_read(ssl, buf, MAXDATASIZE);
  if(DEBUG) printf("Receive Data: %s\n", buf);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  /* send the password followed by a newline */
  if(DEBUG) printf("Sending Data: %s%c", PASSWORD, '\n');
  retcode = SSL_write(ssl, PASSWORD, strlen(PASSWORD));
  retcode = SSL_write(ssl, newline, 1);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  /* Now we take the data out */
  retcode=SSL_read(ssl, buf, MAXDATASIZE);
  if(DEBUG) printf("Receive Data: %s", buf);

  /* Check if we got the "Bad Login" info or the Plugin List start marker */
  if(strstr(buf, PLUGS_START)) login = 1;

  free(buf);
  return login;
}

/* ------------------------------------------------------------------------- *
 * nessus_getplugs tries to retrieve the list of plugins from the server. It *
 * returns the number of retrieved plugins for success, 0 for failure.       *
 * ------------------------------------------------------------------------- */

int nessus_getplugs(){

  char * buf = NULL;
  size_t len;
  int i;

  buf = (char *) malloc(MAXDATASIZE);

  /* cycle through the big chunk of preferences data we */
  /* are receiving.                                     */
 
  while(1) {
    len=SSL_read(ssl, buf, MAXDATASIZE);

    /* check if the end marker hits */
    if(strstr(buf, SERVER_END_MARKER)) { 
      if(DEBUG) printf("Receive Data: %s\n", buf);
      break;
    }

    plugs_counter++;

    //if(DEBUG == 2 ) printf("%s\n", buf);
    if(DEBUG == 2) {
      printf("Len %4.0d Counter %4.0d Plugin ID: ", len, plugs_counter);
      for(i=0;i<5;i++) putchar(buf[i]);
      putchar('\n');
    }

    /* clear the buffer */
    memset(buf, '\0', MAXDATASIZE);
  }

  free(buf);
  return plugs_counter;
}

/* ------------------------------------------------------------------------- *
 * nessus_getprefs retrieves the list of preferences from the server. It     *
 * returns the number of retrieved preferences for success, 0 for failure.   *
 * ------------------------------------------------------------------------- */

int nessus_getprefs() {

  char * buf = NULL;
  size_t len;
  int i = 0;

  buf = (char *) malloc(MAXDATASIZE);

  /* receive the Preferences List start marker */
    len=SSL_read(ssl, buf, strlen(PREFS_START)+1);
  if(DEBUG) printf("Receive Data: %s", buf);

  if(! strstr(buf, PREFS_START)) exit(1);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  while(1) {
    len=SSL_read(ssl, buf, MAXDATASIZE);

    /* check if the end marker hits */
    if(strstr(buf, SERVER_END_MARKER)) {
      if(DEBUG) printf("Receive Data: %s\n", buf);
      break;
    }

    prefs_counter++;

    if(DEBUG == 2) {
      printf("Len %3.0d Counter %3.0d Pref: ", len, prefs_counter);
      if(strlen(buf) < 54) printf("%s", buf);
      else { 
        for(i=0;i<53;i++) putchar(buf[i]);
        putchar('\n');
      }
    }

    /* clear the buffer */
    memset(buf, '\0', MAXDATASIZE);
  }
 
  free(buf);
  return prefs_counter;
}

/* ------------------------------------------------------------------------- *
 * nessus_getrules retrieves the list of rules from the server. It           *
 * returns the number of retrieved rules for success, 0 for failure.         *
 * ------------------------------------------------------------------------- */

int nessus_getrules(){

  char * buf = NULL;
  size_t len;

  buf = (char *) malloc(MAXDATASIZE);

  /* receive the Rules List start marker */
    len=SSL_read(ssl, buf, strlen(RULES_START)+1);
  if(DEBUG) printf("Receive Data: %s", buf);

  if(! strstr(buf, RULES_START)) exit(1);

  while(1) {
    len=SSL_read(ssl, buf, MAXDATASIZE);

    /* check if the end marker hits */
    if(strstr(buf, SERVER_END_MARKER)) { 
      if(DEBUG) printf("Receive Data: %s\n", buf);
      break;
    }

    prefs_counter++;

    if(DEBUG == 2) {
      printf("Len %3.0d Counter %3.0d Rules: ", len, rules_counter);
      printf("%s\n", buf);
    }

    /* clear the buffer */
    memset(buf, '\0', MAXDATASIZE);
  }

  free(buf);
  return rules_counter;
}

int main() {

  int nessusfd;
  int ret;

  nessusfd = nessus_connect(DEST_IP, DEST_PORT);

  ret = nessus_login(USERNAME, PASSWORD);

  if(ret == 0) {
    printf("login of %s with %s failed.\n", USERNAME, PASSWORD);
    close(nessusfd);
    exit(1);
  }

  printf("sucessful login\n");

  plugs_counter = nessus_getplugs();

  if(plugs_counter) printf("Found %d Plugins of 5440.\n", plugs_counter);
  else printf("\nFailed to receive any plugins.\n");

  prefs_counter = nessus_getprefs();

  if(prefs_counter) printf("Found %d Prefs of 165.\n", prefs_counter);
  else printf("\nFailed to receive any preferences.\n");

  rules_counter = nessus_getrules();

  if(rules_counter >= 0) printf("Found %d Rules of 0.\n", rules_counter);
  else printf("\nFailed to receive any Rules.\n");
 
  printf("Closing connection to %s.\n", DEST_IP);
  close(nessusfd);
  exit(0);
}
