/* ------------------------------------------------------------------------ *
 *                                                                          *
 * nessuswc.c  provides functions and routines for various cgi's            *
 *                                                                          *
 * Here I keep functions providing communication with a Nessus daemon.      *
 * They use SSL for encryption (relies on OpenSSL libraries to provide      *
 * the required en/decryption functions). For more info see also the        *
 * demo programs provided in the demo-src directory.                        *
 *                                                                          *
 * It loads the preferences in the according array of structs:              *
 * pluginlist, prefslist, ruleslist                                         *
 *                                                                          *
 * ssl_version=SSLv3 must be set in nessusd.conf.                           *
 * ------------------------------------------------------------------------ */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <openssl/x509v3.h>
#include <openssl/ssl.h>
#include <errno.h>
#include <cgic.h>
#include "nessuswc.h"

const char newline[2]        = {'\n', '\0'};
const char semicolon[2]      = {';', '\0'};
const char whitespace[2]     = {';', '\0'};
      char session_id[24];
      char error_string[255] = "";
      int plugs_counter = 0;
      SSL           *ssl;

/* ------------------------------------------------------------------------- *
 * nessus_connect creates a new tcp session to the nessus server and returns *
 * the file descriptor to it.                                                *
 * ------------------------------------------------------------------------- */
SSL * nessus_connect(char *ip, int port, char *encr, char *cert) {
  int sockfd;
  int retcode;
  int cert_auth = 0;
  struct sockaddr_in dest_addr;
  static SSL_CTX        *ssl_ctx = NULL;
  char 			cacertfile[255] = "";
  char 			clientcertfile[255] = "";

  /* if we get cert=none, we use password authentication */
  if (strcmp(cert, "none")) cert_auth = 1;

  /* create the basic TCP socket */
  sockfd = socket(AF_INET, SOCK_STREAM, 0);

  dest_addr.sin_family=AF_INET;
  dest_addr.sin_port=htons(port);
  dest_addr.sin_addr.s_addr=inet_addr(ip);

  /* Zeroing the rest of the struct */
  memset(&(dest_addr.sin_zero), '\0', 8);

  if ( connect(sockfd, (struct sockaddr *) &dest_addr,
                              sizeof(struct sockaddr)) == -1 ) {
    snprintf(error_string, sizeof(error_string),
             "Error connecting to Nessus server %s on port %d.", ip, port);
    int_error(error_string);
  }

  /* start the SSL initialization */
  ssl = NULL;

  /* initialize SSL library and register algorithms */
  if(SSL_library_init() < 0)
    int_error("Could not initialize the OpenSSL library !");

  /* load  the SSL error messages */
  SSL_load_error_strings();

  /* create a new SSL_CTX object as framework for TLS/SSL enabled functions *
   * use the selected encryption type: TLS1, SSLv2, SSLv3 or auto           *
   * this corresponds to the 'ssl_version' setting in nessusd.conf          */

   if ( strcmp(encr, "TLSv1") == 0 )
      ssl_ctx = SSL_CTX_new(TLSv1_client_method());
   else if ( strcmp(encr, "SSLv2") == 0 )
      ssl_ctx = SSL_CTX_new(SSLv2_client_method());
   else if ( strcmp(encr, "SSLv3") == 0 )
      ssl_ctx = SSL_CTX_new(SSLv3_client_method());
   else if ( strcmp(encr, "Auto") == 0 )
      ssl_ctx = SSL_CTX_new(SSLv23_client_method());
   else int_error("Error unknown SSL encryption method.");

   if ( ssl_ctx == NULL )
      int_error("SSL_CTX_new() context creation error.");

  /* enable all SSL engine bug workaround options (i.e. Netscape, Microsoft) */
  if(SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL) < 0)
    int_error("Error SSL_CTX_set_options.");

  /* choose list of available SSL_CIPHERs with a control string */
  if(! SSL_CTX_set_cipher_list(ssl_ctx, CIPHER_LIST))
    int_error("Error SSL_CTX_set_cipher_listError.");

  /* --- start SSL routines for client cert authentication ----- */
  if (cert_auth) {
    /* verify the CA file and path */
    snprintf(cacertfile, sizeof(cacertfile),
             "%s/%s", CA_CERT_DIR, CA_CERT);
    if(SSL_CTX_load_verify_locations(ssl_ctx, cacertfile, NULL) != 1) {
      snprintf(error_string, sizeof(error_string),
               "Error loading CA certificate: %s.", CA_CERT);
      int_error(error_string);
    }

    /* load the client certificate we present to the nessus server */
    snprintf(clientcertfile, sizeof(clientcertfile),
             "%s/%s", CLIENT_CERT_DIR, cert);
    if(SSL_CTX_use_certificate_chain_file(ssl_ctx, clientcertfile) != 1) {
      snprintf(error_string, sizeof(error_string),
               "Error loading client certificate: %s.\n%s",
                cert, strerror(errno));
      int_error(error_string);
    }

    /* load the client certificate private key */
    if(SSL_CTX_use_PrivateKey_file(ssl_ctx, CLIENT_PRIVKEY,
                                            SSL_FILETYPE_PEM) != 1) {
      snprintf(error_string, sizeof(error_string),
               "Error loading client cert private key: %s.", CLIENT_PRIVKEY);
      int_error(error_string);
    }

    /* Here we set the verification mode to SSL_VERIFY_PEER */
    SSL_CTX_set_verify(ssl_ctx, SSL_VERIFY_PEER, NULL);
    SSL_CTX_set_verify_depth(ssl_ctx, 4);

  /* --- end SSL routines for client cert authentication ----- */
  }

  /* create a new SSL structure for a connection */
  if((ssl = SSL_new(ssl_ctx)) == NULL)
     int_error("Error creating new SSL structure with SSL_new().");

  /* connect the SSL object with the socket file descriptor */
  if(! SSL_set_fd(ssl, sockfd))
     int_error("Error connecting to socket with SSL_set_fd().");

  /* initiate the TLS/SSL handshake with an TLS/SSL server */
  if( (retcode = SSL_connect(ssl)) <= 0) {
    snprintf(error_string, sizeof(error_string),
    "Error SSL_connect() during SSL handshake.\n%d %s", retcode, strerror(SSL_get_error(ssl, retcode)));
    int_error(error_string);
  }

  return ssl;
}

/* ------------------------------------------------------------------------- *
 * nessus_login tries to log in with the username and password provided.     *
 * ------------------------------------------------------------------------- */

void nessus_login(SSL *ssl, char *user, char *pass) {

  char * buf = NULL;
  int retcode = 0;

  /* send the client protocol version, followed by a newline */
  retcode = SSL_write(ssl, CLIENT_NTP_VERSION, strlen(CLIENT_NTP_VERSION));
  retcode = SSL_write(ssl, newline, 1);

  buf = (char *) malloc(MAXDATASIZE);

  /* receive the server protocol version, followed by a newline */

  retcode=SSL_read(ssl, buf, strlen(SERVER_NTP_VERSION)+1);

  /* set global variable server protocol version, remove the newline */
  //strncpy(server_ntp_version, buf, retcode-1);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  /* receive the username prompt */
  retcode=SSL_read(ssl, buf, MAXDATASIZE);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  /* send the username followed by a newline */
  retcode = SSL_write(ssl, user, strlen(user));
  retcode = SSL_write(ssl, newline, 1);

  /* receive the password prompt */
  retcode=SSL_read(ssl, buf, MAXDATASIZE);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  /* send the password, followed by a newline */
  retcode = SSL_write(ssl, pass, strlen(pass));
  retcode = SSL_write(ssl, newline, 1);

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  /* Now we take the data out */
  retcode=SSL_read(ssl, buf, MAXDATASIZE);

  /* Check if we got the "Bad Login" info or the Plugin List start marker */
  if(! strstr(buf, PLUGS_START)) int_error("Can't login to Nessus server.");
  free(buf);
}

/* ------------------------------------------------------------------------- *
 * nessus_getplugs tries to retrieve the list of plugins from the server. It *
 * returns the number of retrieved plugins for success, 0 for failure.       *
 * ------------------------------------------------------------------------- */

int nessus_getplugs(SSL *ssl){

  char * buf = NULL;
  char * buf_ptr = NULL;
  char * tmp_loc = NULL;
  size_t len, diff = 0;

  buf = (char *) malloc(MAXDATASIZE);

  /* cycle through the big chunk of preferences data we */
  /* are receiving.                                     */
 
  while(1) {
    len=SSL_read(ssl, buf, MAXDATASIZE);

    /* check if the end marker hits */
    if(strstr(buf, SERVER_END_MARKER)) break;

    /* cut the plugin in 'buf' into its parts and assign it to a member of  *
     * the 'plugslist' struct array. Parts are separated by string pattern  *   
     * " <|> ", the ";" that needs to be changed into newlines is ignored.  */

    {
      buf_ptr = buf;

      /* collect the plugin_id section */
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(plugslist[plugs_counter].id, buf_ptr,
                  diff - sizeof(PREF_SEP_MARKER));

      /* collect the name section */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(plugslist[plugs_counter].name, buf_ptr,
                    diff - sizeof(PREF_SEP_MARKER));

      /* collect the category section */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(plugslist[plugs_counter].category, buf_ptr,
                    diff - sizeof(PREF_SEP_MARKER));

      /* collect the author/copyright section */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(plugslist[plugs_counter].author, buf_ptr,
                    diff - sizeof(PREF_SEP_MARKER));

      /* collect the descriptive section */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(plugslist[plugs_counter].descr, buf_ptr,
                    diff - sizeof(PREF_SEP_MARKER));

      /* collect the summary section */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(plugslist[plugs_counter].summary, buf_ptr,
                    diff - sizeof(PREF_SEP_MARKER));

      /* collect the family section */
      buf_ptr = tmp_loc - 1;
      diff = (strchr(buf_ptr, '\n') - buf_ptr);
      strncpy(plugslist[plugs_counter].family, buf_ptr, diff);

      /* the last and very important step: save the struct pointer */
      plugslist_ptr[plugs_counter] = &plugslist[plugs_counter];
    }

    plugs_counter++;

    /* clear the buffer */
    memset(buf, '\0', MAXDATASIZE);
  }

  free(buf);
  return plugs_counter;
}

/* ------------------------------------------------------------------------- *
 * nessus_getprefs retrieves the list of preferences from the server. It     *
 * returns the number of retrieved preferences for success, 0 for failure.   *
 * ------------------------------------------------------------------------- */

int nessus_getprefs(SSL *ssl) {

  char * buf = NULL;
  char * buf_ptr = NULL;
  char * tmp_loc = NULL;
  size_t len, diff = 0;
  int	 prefs_counter = 0;

  buf = (char *) malloc(MAXDATASIZE);

  /* receive the Preferences List start marker */
    len=SSL_read(ssl, buf, strlen(PREFS_START)+1);

  if(! strstr(buf, PREFS_START))
   int_error("Server communication error: Expecting preferences start marker.");

  /* clear the buffer */
  memset(buf, '\0', MAXDATASIZE);

  while(1) {
    len=SSL_read(ssl, buf, MAXDATASIZE);

    /* check if the end marker hits */
    if(strstr(buf, SERVER_END_MARKER)) break;

    /* cut the preferences from 'buf' into their parts and assign them    *
     * to be members of the 'prefslist' struct array. Parts are separated *
     * string patterns " <|> ". We ignore other format markers here.      */
    {
      buf_ptr = buf;

      /* collect the pref name section */
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(prefslist[prefs_counter].name, buf_ptr,
                  diff - sizeof(PREF_SEP_MARKER));

      /* collect the pref value section */
      buf_ptr = tmp_loc - 1;
      diff = (strchr(buf_ptr, '\n') - buf_ptr);
      strncpy(prefslist[prefs_counter].value, buf_ptr, diff);

      /* the last and very important step: save the struct pointer */
      prefslist_ptr[prefs_counter] = &prefslist[prefs_counter];
    }

    prefs_counter++;

    /* clear the buffer */
    memset(buf, '\0', MAXDATASIZE);
  }
 
  free(buf);
  return prefs_counter;
}

/* ------------------------------------------------------------------------- *
 * nessus_getrules retrieves the list of rules from the server. It           *
 * returns the number of retrieved rules for success, 0 for failure.         *
 * ------------------------------------------------------------------------- */

int nessus_getrules(SSL *ssl){

  char * buf = NULL;
  size_t len;
  int	 rules_counter = 0;

  buf = (char *) malloc(MAXDATASIZE);

  /* receive the Rules List start marker */
    len=SSL_read(ssl, buf, strlen(RULES_START)+1);

  if(! strstr(buf, RULES_START))
    int_error("Server communication error: Expecting rules start marker.");

  while(1) {
    len=SSL_read(ssl, buf, MAXDATASIZE);

    /* check if the end marker hits */
    if(strstr(buf, SERVER_END_MARKER)) break;

    rules_counter++;

    /* clear the buffer */
    memset(buf, '\0', MAXDATASIZE);
  }

  free(buf);
  return rules_counter;
}

/* ------------------------------------------------------------------------- *
 * nessus_getgroups creates list of categories and families.                 *
 * It also sets the counters for both groups.                                *
 * ------------------------------------------------------------------------- */

void nessus_getgroups(int plugs_counter) {

  int found_catgy;
  int found_famly;
  int i;
  int j;
  int k;
  
  catgy_counter = 0;
  famly_counter = 0;

  /* cycle through all plugins */
  for(i=0; i<plugs_counter; i++) {

    j = 0;
    found_catgy = 0;

    /* cycle through all categories */
    while(j<=catgy_counter) {

      /* check if we already got this category in our list */
      if(strcmp(plugslist_ptr[i]->category, catgylist[j].name) == 0) {

        /* the category exists, add the plugin id to the list ... */
        catgylist[j].plugs_ptr[catgylist[j].plugscount] = plugslist_ptr[i];

        /* ... increase the plugin counter */
        catgylist[j].plugscount++;

        /* ... and set the "category exists" flag */
        found_catgy = 1;
        break;
      }
      j++;
    }
    if(found_catgy == 0) {

      /* the category is a new one, save the category name */
      strcpy(catgylist[catgy_counter].name, plugslist_ptr[i]->category);

      /* .. add the plugin id */
      catgylist[catgy_counter].plugs_ptr[0] = plugslist_ptr[i];

      /* .. set the plugin counter to one for the first plugin found */
      catgylist[catgy_counter].plugscount = 1;

      /* .. set the category "enabled" flag by default to zero */
      catgylist[catgy_counter].enabled = 0;

      /* ... and increase the category counter */
      catgy_counter++;
    }

    k = 0;
    found_famly = 0;

    /* cycle through all families */
    while(k<=famly_counter) {

      if(strcmp(plugslist_ptr[i]->family, famlylist[k].name) == 0) {

        /* the family exists, add the plugin id to the list ... */
        famlylist[k].plugs_ptr[famlylist[k].plugscount] = plugslist_ptr[i];

        /* ... increase the plugin counter */
        famlylist[k].plugscount++;

        /* ... and set the "family exists" flag */
        found_famly = 1;
        break;
      }
      k++;
    }
    if(found_famly == 0) {

      /* the family is a new one, save the family name */
      strcpy(famlylist[famly_counter].name, plugslist_ptr[i]->family);

      /* .. add the plugin id */
      famlylist[famly_counter].plugs_ptr[0] = plugslist_ptr[i];

      /* .. set the plugin counter to one for the first plugin found */
      famlylist[famly_counter].plugscount = 1;

      /* .. set the family "enabled" flag by default to zero */
      famlylist[famly_counter].enabled = 0;

      /* ... and increase the family counter */
      famly_counter++;
    }
  }
}

/* ------------------------------------------------------------------------- *
 * nessus_setprefs sends the list of preferences to the nessus server.       *
 * ------------------------------------------------------------------------- */

void nessus_setprefs(SSL *ssl, int prefs_counter) {

  int retcode = 0;
  int i = 0;

  /* send the client preferences start marker, followed by a newline */
  retcode = SSL_write(ssl, CLIENT_CONF, strlen(CLIENT_CONF));
  retcode = SSL_write(ssl, newline, 1);

  /* return the whole list of server preferences stored in prefslist */

  for (i=0; i<prefs_counter; i++) {

  retcode = SSL_write(ssl, prefslist_ptr[i]->name,
                                     strlen(prefslist_ptr[i]->name));
  retcode = SSL_write(ssl, PREF_SEP_MARKER, strlen(PREF_SEP_MARKER));
  retcode = SSL_write(ssl, prefslist_ptr[i]->value,
                                     strlen(prefslist_ptr[i]->value));
  retcode = SSL_write(ssl, newline, 1);

  }
}

/* ------------------------------------------------------------------------- *
 * nessus_setplugs sends the list of enabled plugins to the nessus server.   *
 * and returns the number of plugin id's that have been send.                *
 * ------------------------------------------------------------------------- */

int nessus_setplugs(SSL *ssl) {

  int retcode = 0;
  int scanplug_counter = 0;
  int i = 0;
  int j = 0;

  //pagehead("debug", 0, cgiOut);
  //fprintf(cgiOut, "<pre>");

  /* send the client plugin list tart marker, !NOT! followed by a newline */
  retcode = SSL_write(ssl, SEND_PLUGID, strlen(SEND_PLUGID));

  /*  send out these plugin id's where the family name has the 'enabled'  *
   *  flag set in 'familylist', separated by a semicolon.                 */

  for(i=0; i<famly_counter; i++) {

    if(famlylist[i].enabled == 1) {

      for(j=0; j<famlylist[i].plugscount; j++) {

        /* send the semicolon, unless we got the first id, in which case we  *
         * don't need to.                                                    */
        if (scanplug_counter != 0)
          retcode = SSL_write(ssl, semicolon, strlen(semicolon));
          // fprintf(cgiOut, "%s", semicolon);

        retcode = SSL_write(ssl, famlylist[i].plugs_ptr[j]->id,
                                   strlen(famlylist[i].plugs_ptr[j]->id));
        // fprintf(cgiOut, "%s", famlylist[i].plugs_ptr[j]->id);

        scanplug_counter++;
      }
    }
  //fprintf(cgiOut, "</pre>");
  }

  retcode = SSL_write(ssl, newline, 1);
  retcode = SSL_write(ssl, CLIENT_END_MARKER, strlen(CLIENT_END_MARKER));
  retcode = SSL_write(ssl, newline, 1);

  return scanplug_counter;
} 

/* ------------------------------------------------------------------------- *
 * nessus_target sends the target ip/netmask to the nessus server.           *
 * ------------------------------------------------------------------------- */

void nessus_target(SSL *ssl, char *target_ip, char *target_mask) {

  int retcode = 0;

  retcode = SSL_write(ssl, CLIENT_ATCK, strlen(CLIENT_ATCK));

  /* here is one of the inconsistencies in the NTP protocol. Sometimes *
   * a whitespace precedes the "<|> CLIENT" marker, sometimes not (see *
   * when the client finishes the preferences sent procedure). Weird!  *
   * Sending a whitespace doesn't help as it will be sent in a packet  *
   * ( ie. like //retcode = SSL_write(ssl, whitespace, 1); )           *
   * of its own and therefore interpreted as a ';' by the server. But  *
   * if the whitespace is missing, the server does not get the "<|>"   *
   * marker right. I worked around by defining the CLIENT_END_MARKER   *
   * with a leading space " " that seems to work even for the prefs    *
   * section.                                                          */

  retcode = SSL_write(ssl, target_ip, strlen(target_ip));
  retcode = SSL_write(ssl, CLIENT_END_MARKER, strlen(CLIENT_END_MARKER));
  retcode = SSL_write(ssl, newline, 1);
}

/* ------------------------------------------------------------------------- *
 * nessus_geterrs gets the scan config errors from the nessus server.        *
 * It returns the number of errors reported, which should be zero.           *
 * ------------------------------------------------------------------------- */

int nessus_geterrs(SSL *ssl) {

  char * buf = NULL;
  int err_counter = 0;
  size_t len;

  buf = (char *) malloc(MAXDATASIZE);

  /* We take the server data out */
  len=SSL_read(ssl, buf, MAXDATASIZE);

  /* Check if we got the "PREFERENCES_ERRORS" list start marker */
  if(! strstr(buf, PREFS_ERROR))
    int_error("Server communication error: Expected preferences error list.");

  while(1) {
    len=SSL_read(ssl, buf, MAXDATASIZE);

    /* check if the end marker hits */
    if(strstr(buf, SERVER_END_MARKER)) break;

    /* not sure how errors look like and if they are separated by    *
     * 'newline' or by the marker ' <|> '. Will see and update code. */
    err_counter++;

    /* clear the buffer */
    memset(buf, '\0', MAXDATASIZE);

  }

  free(buf);
  return err_counter;
}

/* ------------------------------------------------------------------------- *
 * nessus_getstats gets the scan process updates from the nessus server.     *
 * ------------------------------------------------------------------------- */

void nessus_getstats(SSL *ssl, char *session) {

  char * buf = NULL;
  char * buf_ptr = NULL;
  char * tmp_loc = NULL;
  size_t len, diff = 0;
  int stats_counter = 0;

  result_counter = 0;

  buf = (char *) malloc(MAXDATASIZE);

  while(1) {

    /* clear the buffer */
    memset(buf, '\0', MAXDATASIZE);

    len=SSL_read(ssl, buf, MAXDATASIZE);

    /* check if the bye-bye marker hits */
    if(strstr(buf, SERVER_BYE_MARKER)) break;

    /* Check if we got a "SERVER ERROR" start marker */
    if(strstr(buf, ERROR_START)) {
    
      strcpy(reslist[result_counter].service, "n/a");
      buf_ptr = buf + strlen(ERROR_START);
      strncpy(reslist[result_counter].data, buf_ptr, strcspn(buf_ptr, ";"));
      strcpy(reslist[result_counter].plugin_id, "n/a");
      result_counter++;
    }

    /* Check if we got a INFO start marker */
    if(strstr(buf, INFOS_START)) {

      buf_ptr = buf + sizeof(INFOS_START);
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
  
      /* tmp_loc now points at the end of the ip address */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(reslist[result_counter].service, buf_ptr,
                                             diff - sizeof(PREF_SEP_MARKER));
      /* tmp_loc now points at the end of the service */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(reslist[result_counter].data, buf_ptr,
                                             diff - sizeof(PREF_SEP_MARKER));
      /* tmp_loc now points at the end of the data */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(reslist[result_counter].plugin_id, buf_ptr,
                                             diff - sizeof(PREF_SEP_MARKER));
      result_counter++;
    }

    /* Check if we got a HOLE start marker */
    if(strstr(buf, HOLES_START)) {

      buf_ptr = buf + sizeof(HOLES_START);
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
  
      /* tmp_loc now points at the end of the ip address */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(reslist[result_counter].service, buf_ptr,
                                             diff - sizeof(PREF_SEP_MARKER));
      /* tmp_loc now points at the end of the service */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(reslist[result_counter].data, buf_ptr,
                                             diff - sizeof(PREF_SEP_MARKER));
      /* tmp_loc now points at the end of the data */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      strncpy(reslist[result_counter].plugin_id, buf_ptr,
                                             diff - sizeof(PREF_SEP_MARKER));
      result_counter++;
    }

    /* Check if we got a STATUS start marker */
    if(strstr(buf, STATS_START)) {

      buf_ptr = buf + sizeof(STATS_START);
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
  
      /* tmp_loc now points at the end of the ip address */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      // this is the "attack" string
      //strncpy(where???, buf_ptr, diff - sizeof(PREF_SEP_MARKER));
  
      /* tmp_loc now points at the end of the attack string */
      buf_ptr = tmp_loc - 1;
      tmp_loc = strstr(buf_ptr, PREF_SEP_MARKER) + sizeof(PREF_SEP_MARKER);
      diff = tmp_loc - buf_ptr;
      // this is the progress string
      //strncpy(where???, buf_ptr, diff - sizeof(PREF_SEP_MARKER));

      stats_counter++;
    }
  }
}

void write_hostresults(char *t_ip, char * s_ip, char *session) {

  FILE *RESULTS;
  char *outbuf;
  char res_path[127] = "";
  char scantimestr[127] = "";
  char title[72] = "Scan Results for ";
  char table_title[1024] = "";
  char table_title_buf[256] = "";
  struct tm scantime;
  int i,j =0;
  
  strcat(title, t_ip);

  /* convert the timestamp into a human readable string */
  strptime(session, "%s", &scantime);
  strftime(scantimestr, sizeof(scantimestr), "%B, %d. %Y at %T",
                                                             &scantime);

  sprintf(res_path, "%s/session-%s.htm", RESULTS_DIR, session);
  RESULTS = fopen(res_path, "w");
  if(RESULTS == NULL) int_error("Could not start writing the results file.");

  get_dns(t_ip);
  sprintf(table_title_buf, "Scan results for Host %s [%s]", t_ip, dns_name);
  strcat(table_title, table_title_buf);
  get_dns(cgiRemoteAddr);
  sprintf(table_title_buf, " requested by %s [%s]", cgiRemoteAddr, dns_name);
  strcat(table_title, table_title_buf);
  get_dns(s_ip);
  sprintf(table_title_buf, " and executed from Nessus server %s [%s]", s_ip, dns_name);
  strcat(table_title, table_title_buf);
  //
  sprintf(table_title_buf, " on %s.", scantimestr);
  strcat(table_title, table_title_buf);

  pagehead(title, session, RESULTS);
  fprintf(RESULTS, "<p><center>");
  fprintf(RESULTS, "<table border=\"1\" cellspacing=\"0\" ");
  fprintf(RESULTS, "cellpadding=\"2\" width=\"540\">");
  fprintf(RESULTS, "<tr>\n");
  fprintf(RESULTS, "<td bgcolor=\"#000080\" align=\"center\" colspan=3>");
  fprintf(RESULTS, "<font size=\"2\" face=\"Arial\" color=\"#FFFFFF\">");
  fprintf(RESULTS, "%s Scan Results", CLIENT_SW_VERSION);
  fprintf(RESULTS, "</font>");
  fprintf(RESULTS, "</td>\n");
  fprintf(RESULTS, "</tr>");
  fprintf(RESULTS, "<tr>");
  fprintf(RESULTS, "<td bgcolor=\"#8F8F8F\" align=\"left\" colspan=3>");
  fprintf(RESULTS, "<font size=\"2\" face=\"Arial\" color=\"#000080\">");
  fprintf(RESULTS, "%s", table_title);
  fprintf(RESULTS, "</font>");
  fprintf(RESULTS, "</td>\n");
  fprintf(RESULTS, "</tr>");

  for(i=0; i<result_counter; i++) {

    fprintf(RESULTS, "<font size=\"2\" face=\"Arial\">");
    fprintf(RESULTS, "<tr>");
    fprintf(RESULTS, "<td bgcolor=\"#CFCFCF\" align=\"center\"");
    fprintf(RESULTS, " bordercolor=\"#CFCFCF\" width=110>");
    fprintf(RESULTS, "%s", reslist[i].service);
    fprintf(RESULTS, "</td>\n");
    fprintf(RESULTS, "<td bgcolor=\"#CFCFCF\" align=\"center\"");
    fprintf(RESULTS, " bordercolor=\"#CFCFCF\" width=320>");
    for(j=0; j<plugs_counter; j++) {
      if(strcmp(reslist[i].plugin_id, plugslist[j].id) == 0)
        fprintf(RESULTS, "Plugin Name: %s", plugslist[j].name);
    }
    if(strcmp(reslist[i].plugin_id, "n/a") == 0)
      fprintf(RESULTS, "Scanner Error: Can't scan %s", t_ip);

    fprintf(RESULTS, "</td>\n");
    fprintf(RESULTS, "<td bgcolor=\"#CFCFCF\" align=\"center\"");
    fprintf(RESULTS, " bordercolor=\"#CFCFCF\" width=80>");
    fprintf(RESULTS, "ID: %s", reslist[i].plugin_id);
    fprintf(RESULTS, "</td>\n");
    fprintf(RESULTS, "</tr>");

    fprintf(RESULTS, "<tr>");
    fprintf(RESULTS, "<td bgcolor=\"#FFFFFF\" align=\"left\" colspan=3>");

    if((outbuf = strtok(reslist[i].data, ";"))) {
      fprintf(RESULTS, "%s<br>", outbuf);
      while((outbuf = strtok(NULL, ";"))) {
        fprintf(RESULTS, "%s<br>", outbuf);
      }
    }
    fprintf(RESULTS, "</td>\n");
    fprintf(RESULTS, "</tr>");
    fprintf(RESULTS, "</font>");
  }

  fprintf(RESULTS, "<tr>");
  fprintf(RESULTS, "<td bgcolor=\"#000080\" align=\"center\" colspan=3>");
  fprintf(RESULTS, "<font size=\"2\" face=\"Arial\" color=\"#FFFFFF\">");
  fprintf(RESULTS, "End");
  fprintf(RESULTS, "</font>");
  fprintf(RESULTS, "</td>\n");
  fprintf(RESULTS, "</tr>");
  fprintf(RESULTS, "</table>\n");
  fprintf(RESULTS, "</BODY>");
  fclose(RESULTS);
}

char * create_session_id(char *target_ip) {

int timestamp = 0;
int decimal_ip = 0;

  timestamp = time(NULL);
  decimal_ip = inet_addr(target_ip);
  snprintf(session_id, sizeof(session_id), "%d_%d", timestamp, decimal_ip);

  return session_id;
}

void get_dns(char *ip) {

  struct hostent *host;
  struct in_addr *my_addr;

  strcpy(dns_name, "");
  my_addr=(struct in_addr*)malloc(sizeof(struct in_addr));
  my_addr->s_addr=inet_addr(ip);
  host = gethostbyaddr((char *) my_addr, 4, AF_INET);
  if(host != NULL) strncpy(dns_name, host->h_name, sizeof(dns_name));
  else strncpy(dns_name, "unknown", sizeof(dns_name));
}

